import logging
logger = logging.getLogger(__name__)

__version__ = '0.3.4'


def main(global_config, **settings):
    """
    This function returns a Pyramid WSGI application.
    """
    from pyramid.config import Configurator

    config = Configurator(settings=settings)

    # include twitcher components
    config.include('twitcher.config')
    config.include('twitcher.frontpage')
    config.include('twitcher.rpcinterface')
    config.include('twitcher.owsproxy')
    config.include('twitcher.wps')

    # tweens/middleware
    # TODO: maybe add tween for exception handling or use unknown_failure view
    config.include('twitcher.tweens')

    config.scan()

    return config.make_wsgi_app()



import pytest
import unittest
import mock
from tempfile import mkdtemp

from twitcher.esgf import ESGFAccessManager


class ESGFTestCase(unittest.TestCase):
    def setUp(self):
        self.mgr = ESGFAccessManager(
            slcs_service_url="https://localhost:5000",
            base_dir=mkdtemp())
        self.mgr._retrieve_certificate = mock.MagicMock(return_value=True)

    def test_logon_with_token(self):
        assert self.mgr.logon(access_token="abcdef") is True



from pyramid.settings import asbool
from pyramid.tweens import EXCVIEW

from twitcher.owsexceptions import OWSException, OWSNoApplicableCode
from twitcher.owssecurity import owssecurity_factory

import logging
logger = logging.getLogger(__name__)


def includeme(config):
    settings = config.registry.settings

    if asbool(settings.get('twitcher.ows_security', True)):
        logger.info('Add OWS security tween')
        config.add_tween(OWS_SECURITY, under=EXCVIEW)


def ows_security_tween_factory(handler, registry):
    """A tween factory which produces a tween which raises an exception
    if access to OWS service is not allowed."""

    security = owssecurity_factory(registry)

    def ows_security_tween(request):
        try:
            security.check_request(request)
            return handler(request)
        except OWSException as err:
            logger.exception("security check failed.")
            return err
        except Exception as err:
            logger.exception("unknown error")
            return OWSNoApplicableCode(err.message)

    return ows_security_tween


OWS_SECURITY = 'twitcher.tweens.ows_security_tween_factory'



import os
import tempfile

from pyramid.settings import asbool


import logging
LOGGER = logging.getLogger("TWITCHER")


def _workdir(request):
    settings = request.registry.settings
    workdir = settings.get('twitcher.workdir')
    workdir = workdir or tempfile.gettempdir()
    if not os.path.exists(workdir):
        os.makedirs(workdir)
    LOGGER.debug('using workdir %s', workdir)
    return workdir


def _prefix(request):
    settings = request.registry.settings
    prefix = settings.get('twitcher.prefix')
    prefix = prefix or 'twitcher_'
    return prefix


def includeme(config):
    # settings = config.registry.settings

    LOGGER.debug("Loading twitcher configuration.")

    config.add_request_method(_workdir, 'workdir', reify=True)
    config.add_request_method(_prefix, 'prefix', reify=True)



import pytest
import unittest
import mock

from pyramid import testing
from pyramid.testing import DummyRequest

from twitcher.owsrequest import OWSRequest
from twitcher.owsexceptions import OWSInvalidParameterValue, OWSMissingParameterValue


class OWSRequestWpsTestCase(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp()

        
    def tearDown(self):
        testing.tearDown()


    def test_get_getcaps_request(self):
        params = dict(request="GetCapabilities", service="WPS")
        request = DummyRequest(params=params)
        ows_req = OWSRequest(request)
        assert ows_req.request == 'getcapabilities'
        assert ows_req.service == 'wps'

        
    def test_get_describeprocess_request(self):
        params = dict(request="DescribeProcess", service="wps", version="1.0.0")
        request = DummyRequest(params=params)
        ows_req = OWSRequest(request)
        assert ows_req.request == 'describeprocess'
        assert ows_req.service == 'wps'
        assert ows_req.version == '1.0.0'


    def test_get_execute_request(self):
        params = dict(request="execute", service="Wps", version="1.0.0")
        request = DummyRequest(params=params)
        ows_req = OWSRequest(request)
        assert ows_req.request == 'execute'
        assert ows_req.service == 'wps'
        assert ows_req.version == '1.0.0'


    def test_get_false_request(self):
        params = dict(request="tellmemore", service="Wps", version="1.0.0")
        request = DummyRequest(params=params)
        with pytest.raises(OWSInvalidParameterValue) as e_info:
            ows_req = OWSRequest(request)

            
    def test_get_missing_request(self):
        params = dict(service="wps", version="1.0.0")
        request = DummyRequest(params=params)
        with pytest.raises(OWSMissingParameterValue) as e_info:
            ows_req = OWSRequest(request)

            
    def test_get_false_service(self):
        params = dict(request="execute", service="ATM", version="1.0.0")
        request = DummyRequest(params=params)
        with pytest.raises(OWSInvalidParameterValue) as e_info:
            ows_req = OWSRequest(request)


    def test_get_missing_service(self):
        params = dict(request="Execute", version="1.0.0")
        request = DummyRequest(params=params)
        with pytest.raises(OWSMissingParameterValue) as e_info:
            ows_req = OWSRequest(request)


    def test_post_getcaps_request(self):
        request = DummyRequest(post={})
        request.body = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <GetCapabilities service="WPS" acceptVersions="1.0.0" language="en-CA"/>"""
        ows_req = OWSRequest(request)
        assert ows_req.request == 'getcapabilities'
        assert ows_req.service == 'wps'


    def test_post_false_request(self):
        request = DummyRequest(post={})
        request.body = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <MyCaps service="WPS" acceptVersions="1.0.0" language="en-CA"/>"""
        with pytest.raises(OWSInvalidParameterValue) as e_info:
            ows_req = OWSRequest(request)


    def test_post_false_service(self):
        request = DummyRequest(post={})
        request.body = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <GetCapabilities service="ATM" acceptVersions="1.0.0" language="en-CA"/>"""
        with pytest.raises(OWSInvalidParameterValue) as e_info:
            ows_req = OWSRequest(request)

        
    def test_post_describeprocess_request(self):
        request = DummyRequest(post={})
        request.body = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <DescribeProcess service="WPS" version="1.0.0" language="en" xmlns:ows="http://www.opengis.net/ows/1.1">
          <ows:Identifier>intersection</ows:Identifier>
          <ows:Identifier>union</ows:Identifier>
        </DescribeProcess>""" 
        ows_req = OWSRequest(request)
        assert ows_req.request == 'describeprocess'
        assert ows_req.service == 'wps'
        assert ows_req.version == '1.0.0'
        

    def test_post_execute_request(self):
        request = DummyRequest(post={})
        request.body = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <wps:Execute service="WPS" version="1.0.0" xmlns:wps="http://www.opengis.net/wps/1.0.0" xmlns:ows="http://www.opengis.net/ows/1.1" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.opengis.net/wps/1.0.0/../wpsExecute_request.xsd">
    <ows:Identifier>Buffer</ows:Identifier>
    <wps:DataInputs>
        <wps:Input>
            <ows:Identifier>InputPolygon</ows:Identifier>
            <ows:Title>Playground area</ows:Title>
            <wps:Reference xlink:href="http://foo.bar/some_WFS_request.xml"/>
        </wps:Input>
        <wps:Input>
            <ows:Identifier>BufferDistance</ows:Identifier>
            <ows:Title>Distance which people will walk to get to a playground.</ows:Title>
            <wps:Data>
                <wps:LiteralData>400</wps:LiteralData>
            </wps:Data>
        </wps:Input>
    </wps:DataInputs>
    <wps:ResponseForm>
        <wps:ResponseDocument storeExecuteResponse="true">
            <wps:Output asReference="true">
                <ows:Identifier>BufferedPolygon</ows:Identifier>
                <ows:Title>Area serviced by playground.</ows:Title>
                <ows:Abstract>Area within which most users of this playground will live.</ows:Abstract>
            </wps:Output>
        </wps:ResponseDocument>
    </wps:ResponseForm>
</wps:Execute>"""
        ows_req = OWSRequest(request)
        assert ows_req.request == 'execute'
        assert ows_req.service == 'wps'
        assert ows_req.version == '1.0.0'

        



import pytest
import unittest
import webtest
from pyramid import testing
from .common import setup_with_mongodb

from twitcher.store import servicestore_factory


class OWSProxyAppTest(unittest.TestCase):

    def setUp(self):
        config = setup_with_mongodb()
        self._setup_registry(config)
        config.include('twitcher.owsproxy')
        config.include('twitcher.tweens')
        self.app = webtest.TestApp(config.make_wsgi_app())

    def tearDown(self):
        testing.tearDown()

    def _setup_registry(self, config):
        registry = servicestore_factory(config.registry)
        registry.clear_services()
        # TODO: testing against ourselfs ... not so good
        url = "https://localhost:5000/ows/wps"
        registry.register_service(url=url, name="twitcher")

    @pytest.mark.skip(reason="no way of currently testing this")
    @pytest.mark.online
    def test_getcaps(self):
        resp = self.app.get('/ows/proxy/twitcher?service=wps&request=getcapabilities')
        assert resp.status_code == 200
        assert resp.content_type == 'text/xml'
        resp.mustcontain('</wps:Capabilities>')

    @pytest.mark.skip(reason="no way of currently testing this")
    @pytest.mark.online
    def test_describeprocess(self):
        resp = self.app.get(
            '/ows/proxy/twitcher?service=wps&request=describeprocess&version=1.0.0&identifier=dummyprocess')
        assert resp.status_code == 200
        assert resp.content_type == 'text/xml'
        resp.mustcontain('</wps:ProcessDescriptions>')

    @pytest.mark.skip(reason="no way of currently testing this")
    @pytest.mark.online
    def test_execute_not_allowed(self):
        resp = self.app.get('/ows/proxy/twitcher?service=wps&request=execute&version=1.0.0&identifier=dummyprocess')
        assert resp.status_code == 200
        assert resp.content_type == 'text/xml'
        print resp.body
        resp.mustcontain('<Exception exceptionCode="NoApplicableCode" locator="AccessForbidden">')



"""
Based on unitests in https://github.com/wndhydrnt/python-oauth2/tree/master/oauth2/test
"""

import pytest
import unittest
import mock

from twitcher.datatype import AccessToken
from twitcher.utils import expires_at
from twitcher.store.mongodb import MongodbTokenStore


class MongodbTokenStoreTestCase(unittest.TestCase):
    def setUp(self):
        self.access_token = AccessToken(token="abcdef", expires_at=expires_at(hours=1))

    def test_fetch_by_token(self):
        collection_mock = mock.Mock(spec=["find_one"])
        collection_mock.find_one.return_value = self.access_token

        store = MongodbTokenStore(collection=collection_mock)
        access_token = store.fetch_by_token(token=self.access_token.token)

        collection_mock.find_one.assert_called_with({"token": self.access_token.token})
        assert isinstance(access_token, AccessToken)

    def test_save_token(self):
        collection_mock = mock.Mock(spec=["insert_one"])

        store = MongodbTokenStore(collection=collection_mock)
        store.save_token(self.access_token)

        collection_mock.insert_one.assert_called_with(self.access_token)

from twitcher.datatype import Service
from twitcher.store.mongodb import MongodbServiceStore


class MongodbServiceStoreTestCase(unittest.TestCase):
    def setUp(self):
        self.service = dict(name="loving_flamingo", url="http://somewhere.over.the/ocean", type="wps",
                            public=False, c4i=False)
        self.service_public = dict(name="open_pingu", url="http://somewhere.in.the/deep_ocean", type="wps",
                                   public=True, c4i=False)
        self.service_special = dict(url="http://wonderload", name="A special Name", type='wps')

    def test_fetch_by_name(self):
        collection_mock = mock.Mock(spec=["find_one"])
        collection_mock.find_one.return_value = self.service

        store = MongodbServiceStore(collection=collection_mock)
        service = store.fetch_by_name(name=self.service['name'])

        collection_mock.find_one.assert_called_with({"name": self.service['name']})
        assert isinstance(service, dict)

    def test_save_service_default(self):
        collection_mock = mock.Mock(spec=["insert_one", "find_one", "count"])
        collection_mock.count.return_value = 0
        collection_mock.find_one.return_value = self.service

        store = MongodbServiceStore(collection=collection_mock)
        store.save_service(Service(self.service))

        collection_mock.insert_one.assert_called_with(self.service)

    def test_save_service_with_special_name(self):
        collection_mock = mock.Mock(spec=["insert_one", "find_one", "count"])
        collection_mock.count.return_value = 0
        collection_mock.find_one.return_value = self.service_special

        store = MongodbServiceStore(collection=collection_mock)
        store.save_service(Service(self.service_special))

        collection_mock.insert_one.assert_called_with({
            'url': 'http://wonderload', 'type': 'wps', 'name': 'a_special_name', 'public': False, 'c4i': False})

    def test_save_service_public(self):
        collection_mock = mock.Mock(spec=["insert_one", "find_one", "count"])
        collection_mock.count.return_value = 0
        collection_mock.find_one.return_value = self.service_public

        store = MongodbServiceStore(collection=collection_mock)
        store.save_service(Service(self.service_public))

        collection_mock.insert_one.assert_called_with(self.service_public)



"""
Testing the Twithcer XML-RPC interface.
"""
import pytest
import unittest
import webtest
import pyramid.testing

from twitcher._compat import PY2
from twitcher._compat import xmlrpclib

from twitcher.tests.functional.common import setup_with_mongodb
from twitcher.tests. functional.common import setup_mongodb_tokenstore
from twitcher.tests.functional.common import setup_mongodb_servicestore


class XMLRPCInterfaceAppTest(unittest.TestCase):

    def setUp(self):
        config = setup_with_mongodb()
        self.token = setup_mongodb_tokenstore(config)
        setup_mongodb_servicestore(config)
        config.include('twitcher.rpcinterface')
        self.app = webtest.TestApp(config.make_wsgi_app())

    def tearDown(self):
        pyramid.testing.tearDown()

    def _callFUT(self, method, params):
        if PY2:
            xml = xmlrpclib.dumps(params, methodname=method)
        else:
            xml = xmlrpclib.dumps(params, methodname=method).encode('utf-8')
        print xml
        resp = self.app.post('/RPC2', content_type='text/xml', params=xml)
        assert resp.status_int == 200
        assert resp.content_type == 'text/xml'
        print resp.body
        return xmlrpclib.loads(resp.body)[0][0]

    @pytest.mark.online
    def test_generate_token_and_revoke_it(self):
        # gentoken
        resp = self._callFUT('generate_token', (1, {}))
        assert 'access_token' in resp
        assert 'expires_at' in resp
        # revoke
        resp = self._callFUT('revoke_token', (resp['access_token'],))
        assert resp is True
        # revoke all
        resp = self._callFUT('revoke_all_tokens', ())
        assert resp is True

    @pytest.mark.online
    def test_register_service_and_unregister_it(self):
        service = {'url': 'http://localhost/wps', 'name': 'test_emu',
                   'type': 'wps', 'public': False, 'c4i': False}
        # register
        resp = self._callFUT('register_service', (
            service['url'],
            service,
            False))
        assert resp == service

        # get by name
        resp = self._callFUT('get_service_by_name', (service['name'],))
        assert resp == service

        # get by url
        resp = self._callFUT('get_service_by_url', (service['url'],))
        assert resp == service

        # list
        resp = self._callFUT('list_services', ())
        assert resp == [service]

        # unregister
        resp = self._callFUT('unregister_service', (service['name'],))
        assert resp is True

        # clear
        resp = self._callFUT('clear_services', ())
        assert resp is True



"""
generates a nice birdy name.

This module is inspired by
`docker names-generator.go <https://github.com/docker/docker/blob/master/pkg/namesgenerator/names-generator.go>`_
"""

import random
import re

left = ["admiring",
        "adoring",
        "agitated",
        "amazing",
        "angry",
        "awesome",
        "backstabbing",
        "berserk",
        "big",
        "blind",
        "boring",
        "clever",
        "cocky",
        "compassionate",
        "condescending",
        "cranky",
        "desperate",
        "determined",
        "distracted",
        "dreamy",
        "drunk",
        "ecstatic",
        "elated",
        "elegant",
        "evil",
        "fervent",
        "focused",
        "furious",
        "gigantic",
        "gloomy",
        "goofy",
        "grave",
        "happy",
        "high",
        "hopeful",
        "hungry",
        "insane",
        "jolly",
        "jovial",
        "kickass",
        "lonely",
        "loving",
        "mad",
        "modest",
        "naughty",
        "nostalgic",
        "pensive",
        "prickly",
        "reverent",
        "romantic",
        "running",
        "sad",
        "serene",
        "sharp",
        "sick",
        "silly",
        "sleepy",
        "small",
        "stoic",
        "stupefied",
        "suspicious",
        "tender",
        "thirsty",
        "tiny",
        "trusting",
        ]

# pick a nice bird: https://en.wikipedia.org/wiki/List_of_birds_by_common_name

right = [
    "albatrosse",
    "antbird",
    "buzzard",
    "cassowary",
    "catbird",
    "chicken",
    "crane",
    "cuckoo",

    # dodrio: http://bulbapedia.bulbagarden.net/wiki/Dodrio_%28Pok%C3%A9mon%29
    "dodrio",

    "dove",
    "duck",
    "eagle",
    "emu",
    "figbird",
    "flamingo",
    "flycatcher",
    "goldfinch",
    "goose",
    "grouse",
    "hawk",
    "honeyeater",
    "hornbill",
    "hummingbird",
    "ibis",
    "kingfisher",
    "kiwi",
    "leafbird",
    "lovebird",
    "malleefowl",
    "mockingbird",
    "mousebird",
    "nightjar",
    "ostriche",
    "owl",
    "parrot",
    "pelican",
    "penguin",
    "pheasant",
    "pigeon",
    "roadrunner",
    "satinbird",
    "seagull",
    "snowfinch",
    "sparrow",
    "starling",
    "stork",
    "sugarbird",
    "sunbird",
    "swan",
    "swift",
    "tinamou",
    "toucan",
    "trogon",
    "turaco",
    "turkey",
    "woodcreeper",
    "woodpecker",
]


def get_random_name(retry=False):
    """
    generates a random name from the list of adjectives and birds in this package
    formatted as "adjective_surname". For example 'loving_sugarbird'. If retry is non-zero, a random
    integer between 0 and 100 will be added to the end of the name, e.g `loving_sugarbird3`
    """
    name = "%s_%s" % (left[random.randint(0, len(left) - 1)], right[random.randint(0, len(right) - 1)])
    if retry is True:
        name = "%s%d" % (name, random.randint(0, 100))
    return name


def get_sane_name(name, minlen=3, maxlen=25):
    if name is None or len(name.strip()) < minlen:
        return None
    return re.sub("[^a-z]", "_", name.strip().lower()[:maxlen])



from pyramid import testing

from twitcher.tokengenerator import tokengenerator_factory
from twitcher.store import tokenstore_factory
from twitcher.store import servicestore_factory


def setup_with_mongodb():
    settings = {'mongodb.host': '127.0.0.1', 'mongodb.port': '27027', 'mongodb.db_name': 'twitcher_test'}
    config = testing.setUp(settings=settings)
    return config


def setup_mongodb_tokenstore(config):
    store = tokenstore_factory(config.registry)
    generator = tokengenerator_factory(config.registry)
    store.clear_tokens()
    access_token = generator.create_access_token()
    store.save_token(access_token)
    return access_token.token


def setup_mongodb_servicestore(config):
    store = servicestore_factory(config.registry)
    store.clear_services()



"""
pywps 4.x wrapper
"""
import os

from pyramid.response import Response
from pyramid.wsgi import wsgiapp2
from pyramid.httpexceptions import HTTPBadRequest
from pyramid.settings import asbool, aslist

from twitcher.processes import processes
from twitcher.owsexceptions import OWSNoApplicableCode


import logging
LOGGER = logging.getLogger(__name__)

PYWPS_CFG = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'wps.cfg')


def _wps_cfg(request):
    settings = request.registry.settings
    return settings.get('twitcher.wps_cfg')


@wsgiapp2
def pywps_view(environ, start_response):
    """
    * TODO: add xml response renderer
    * TODO: fix exceptions ... use OWSException (raise ...)
    """
    from pywps.app.Service import Service
    LOGGER.debug('pywps env: %s', environ.keys())

    # call pywps application
    if 'PYWPS_CFG' not in environ:
        environ['PYWPS_CFG'] = PYWPS_CFG
    service = Service(processes, [environ['PYWPS_CFG']])
    return service(environ, start_response)


def includeme(config):
    settings = config.registry.settings

    if asbool(settings.get('twitcher.wps', True)):
        LOGGER.debug("Twitcher WPS enabled.")

        # include twitcher config
        config.include('twitcher.config')

        config.add_route('wps', '/ows/wps')
        config.add_route('wps_secured', '/ows/wps/{access_token}')
        config.add_view(pywps_view, route_name='wps')
        config.add_view(pywps_view, route_name='wps_secured')
        config.add_request_method(_wps_cfg, 'wps_cfg', reify=True)



import os

RESOURCES_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), 'resources'))

WPS_CAPS_EMU_XML = os.path.join(RESOURCES_PATH, 'wps_caps_emu.xml')
WMS_CAPS_NCWMS2_111_XML = os.path.join(RESOURCES_PATH, 'wms_caps_ncwms2_111.xml')
WMS_CAPS_NCWMS2_130_XML = os.path.join(RESOURCES_PATH, 'wms_caps_ncwms2_130.xml')



# MongoDB
# http://docs.pylonsproject.org/projects/pyramid-cookbook/en/latest/database/mongodb.html
# maybe use event to register mongodb

import pymongo

import logging
logger = logging.getLogger(__name__)


def mongodb(registry):
    settings = registry.settings
    client = pymongo.MongoClient(settings['mongodb.host'], int(settings['mongodb.port']))
    db = client[settings['mongodb.db_name']]
    db.services.create_index("name", unique=True)
    db.services.create_index("url", unique=True)
    return db


def includeme(config):
    config.registry.db = mongodb(config.registry)

    def _add_db(request):
        db = request.registry.db
        # if db_url.username and db_url.password:
        #     db.authenticate(db_url.username, db_url.password)
        return db
    config.add_request_method(_add_db, 'db', reify=True)



"""
Based on unitests in https://github.com/wndhydrnt/python-oauth2/tree/master/oauth2/test
"""

import pytest
import unittest
import mock

from twitcher.tokengenerator import UuidTokenGenerator


class UuidTokenGeneratorTestCase(unittest.TestCase):
    def setUp(self):
        self.generator = UuidTokenGenerator()

    def test_generate(self):
        token = self.generator.generate()
        assert len(token) == 32

    def test_create_access_token_default(self):
        access_token = self.generator.create_access_token()
        assert len(access_token.token) == 32
        assert access_token.expires_in <= 3600

    def test_create_access_non_default_hours(self):
        access_token = self.generator.create_access_token(valid_in_hours=2)
        assert len(access_token.token) == 32
        assert access_token.expires_in <= 3600 * 2



"""
This python 2.x/3.x compatibility modules is based on the pywps 4.x code.
"""

import logging
import sys


LOGGER = logging.getLogger('twitcher')
PY2 = sys.version_info[0] == 2
PY3 = not PY2

if PY2:
    LOGGER.debug('Python 2.x')
    text_type = unicode  # noqa
    from StringIO import StringIO
    # from flufl.enum import Enum
    from urlparse import urlparse
    from urlparse import urljoin
    from urllib2 import urlopen
    import xmlrpclib
else:
    LOGGER.debug('Python 3.x')
    text_type = str
    from io import StringIO
    # from enum import Enum
    from urllib.parse import urlparse
    from urllib.parse import urljoin
    from urllib.request import urlopen
    import xmlrpc.client as xmlrpclib



"""
Testing the Twitcher API.
"""
import pytest
import unittest

from twitcher.api import TokenManager
from twitcher.tokengenerator import UuidTokenGenerator
from twitcher.store.memory import MemoryTokenStore


class TokenManagerTest(unittest.TestCase):

    def setUp(self):
        self.tokenmgr = TokenManager(
            tokengenerator=UuidTokenGenerator(),
            tokenstore=MemoryTokenStore()
        )

    def test_generate_token_and_revoke_it(self):
        # gentoken
        resp = self.tokenmgr.generate_token()
        assert 'access_token' in resp
        assert 'expires_at' in resp
        # revoke
        resp = self.tokenmgr.revoke_token(resp['access_token'])
        assert resp is True
        # revoke all
        resp = self.tokenmgr.revoke_all_tokens()
        assert resp is True

    def test_generate_token_with_data(self):
        # gentoken
        resp = self.tokenmgr.generate_token(valid_in_hours=1, data={'esgf_token': 'abcdef'})
        assert 'access_token' in resp
        assert 'expires_at' in resp
        # check data
        access_token = self.tokenmgr.store.fetch_by_token(resp['access_token'])
        assert access_token.data == {'esgf_token': 'abcdef'}


from twitcher.api import Registry
from twitcher.store.memory import MemoryServiceStore


class RegistryTest(unittest.TestCase):

    def setUp(self):
        self.reg = Registry(servicestore=MemoryServiceStore())

    def test_register_service_and_unregister_it(self):
        service = {'url': 'http://localhost/wps', 'name': 'test_emu',
                   'type': 'wps', 'public': False, 'c4i': False}
        # register
        resp = self.reg.register_service(
            service['url'],
            service,
            False)
        assert resp == service

        # get by name
        resp = self.reg.get_service_by_name(service['name'])
        assert resp == service

        # get by url
        resp = self.reg.get_service_by_url(service['url'])
        assert resp == service

        # list
        resp = self.reg.list_services()
        assert resp == [service]

        # unregister
        resp = self.reg.unregister_service(service['name'])
        assert resp is True

        # clear
        resp = self.reg.clear_services()
        assert resp is True



import unittest

from twitcher.datatype import AccessToken
from twitcher.store.memory import MemoryTokenStore


class MemoryTokenStoreTestCase(unittest.TestCase):
    def setUp(self):
        self.access_token_data = {"token": "xyz",
                                  "data": {"name": "test"},
                                  }
        self.test_store = MemoryTokenStore()

    def test_save_token_and_fetch_by_token(self):
        access_token = AccessToken(**self.access_token_data)

        assert self.test_store.save_token(access_token)
        assert self.test_store.fetch_by_token(access_token.token) == access_token


from twitcher.datatype import Service
from twitcher.store.memory import MemoryServiceStore


class MemoryServiceStoreTestCase(unittest.TestCase):
    def setUp(self):
        self.service_data = {'url': 'http://localhost:8094/wps',
                             'name': 'emu',
                             'c4i': False,
                             'public': False,
                             'type': 'WPS',
                             }
        self.test_store = MemoryServiceStore()

    def test_save_service_and_fetch_service(self):
        service = Service(**self.service_data)

        assert self.test_store.save_service(service)
        assert self.test_store.fetch_by_url(service.url) == service
        assert self.test_store.fetch_by_name(service.name) == service



import ssl
from datetime import datetime

from twitcher._compat import urlparse
from twitcher._compat import xmlrpclib

import logging
LOGGER = logging.getLogger("TWITCHER")


def _create_https_context(verify=True):
    context = ssl._create_default_https_context()
    if verify is False:
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
    return context


def _create_server(url, username=None, password=None, verify=True):
    # TODO: disable basicauth when username is not set
    username = username or 'nouser'
    password = password or 'nopass'

    parsed = urlparse(url)
    url = "%s://%s:%s@%s%s" % (parsed.scheme, username, password, parsed.netloc, parsed.path)
    context = _create_https_context(verify=verify)
    server = xmlrpclib.ServerProxy(url, context=context)
    return server


def xmlrpc_error_handler(wrapped):
    def _handle_error(*args, **kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except xmlrpclib.Fault as e:
            LOGGER.error("A fault occurred: %s (%d)", e.faultString, e.faultCode)
            raise
        except xmlrpclib.ProtocolError as e:
            LOGGER.error(
                "A protocol error occurred. URL: %s, HTTP/HTTPS headers: %s, Error code: %d, Error message: %s",
                e.url, e.headers, e.errcode, e.errmsg)
            raise
        except xmlrpclib.ResponseError as e:
            LOGGER.error(
                "A response error occured. Maybe service needs authentication with username and password? %s",
                e.message)
            raise
        except Exception as e:
            LOGGER.error(
                " Unknown error occured. "
                "Maybe you need to use the \"--insecure\" option to access the service on HTTPS? "
                "Is your service running and did you specify the correct service url (port)? "
                "%s",
                e.message)
            raise
        else:
            return result
    return _handle_error


class TwitcherService(object):
    def __init__(self, url, username=None, password=None, verify=True):
        self.server = _create_server(url, username=username, password=password, verify=verify)

    # tokens

    @xmlrpc_error_handler
    def generate_token(self, valid_in_hours=1, data=None):
        data = data or {}
        return self.server.generate_token(valid_in_hours, data)

    @xmlrpc_error_handler
    def revoke_token(self, token):
        return self.server.revoke_token(token)

    @xmlrpc_error_handler
    def revoke_all_tokens(self):
        return self.server.revoke_all_tokens()

    # service registry

    @xmlrpc_error_handler
    def register_service(self, url, data=None, overwrite=True):
        data = data or {}
        return self.server.register_service(url, data, overwrite)

    @xmlrpc_error_handler
    def unregister_service(self, name):
        return self.server.unregister_service(name)

    @xmlrpc_error_handler
    def list_services(self):
        return self.server.list_services()

    @xmlrpc_error_handler
    def clear_services(self):
        return self.server.clear_services()

    @xmlrpc_error_handler
    def get_service_by_url(self, url):
        return self.server.get_service_by_url(url)

    @xmlrpc_error_handler
    def get_service_by_name(self, name):
        return self.server.get_service_by_name(name)



import pytest
from lxml import etree
from twitcher import utils
from twitcher.exceptions import ServiceNotFound
from twitcher.tests.common import WPS_CAPS_EMU_XML, WMS_CAPS_NCWMS2_111_XML, WMS_CAPS_NCWMS2_130_XML


def test_is_url_valid():
    assert utils.is_valid_url("http://somewhere.org") is True
    assert utils.is_valid_url("https://somewhere.org/my/path") is True
    assert utils.is_valid_url("file:///my/path") is True
    assert utils.is_valid_url("/my/path") is False
    assert utils.is_valid_url(None) is False


def test_parse_service_name():
    assert 'emu' == utils.parse_service_name("/ows/proxy/emu")
    assert 'emu' == utils.parse_service_name("/ows/proxy/emu/foo/bar")
    assert 'emu' == utils.parse_service_name("/ows/proxy/emu/")
    with pytest.raises(ServiceNotFound) as e_info:
        assert 'emu' == utils.parse_service_name("/ows/proxy/")
    with pytest.raises(ServiceNotFound) as e_info:
        assert 'emu' == utils.parse_service_name("/ows/nowhere/emu")


def test_baseurl():
    assert utils.baseurl('http://localhost:8094/wps') == 'http://localhost:8094/wps'
    assert utils.baseurl('http://localhost:8094/wps?service=wps&request=getcapabilities') == 'http://localhost:8094/wps'
    assert utils.baseurl('https://localhost:8094/wps?service=wps&request=getcapabilities') ==\
        'https://localhost:8094/wps'
    with pytest.raises(ValueError) as e_info:
        utils.baseurl('ftp://localhost:8094/wps')


def test_path_elements():
    assert utils.path_elements('/ows/proxy/lovely_bird') == ['ows', 'proxy', 'lovely_bird']
    assert utils.path_elements('/ows/proxy/lovely_bird/') == ['ows', 'proxy', 'lovely_bird']
    assert utils.path_elements('/ows/proxy/lovely_bird/ ') == ['ows', 'proxy', 'lovely_bird']


def test_lxml_strip_ns():
    import lxml.etree
    wpsxml = """
<wps100:Execute
xmlns:wps100="http://www.opengis.net/wps/1.0.0"
xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
service="WPS"
version="1.0.0"
xsi:schemaLocation="http://www.opengis.net/wps/1.0.0 http://schemas.opengis.net/wps/1.0.0/wpsExecute_request.xsd"/>"""

    doc = lxml.etree.fromstring(wpsxml)
    assert doc.tag == '{http://www.opengis.net/wps/1.0.0}Execute'
    utils.lxml_strip_ns(doc)
    assert doc.tag == 'Execute'


def test_replace_caps_url_wps():
    doc = etree.parse(WPS_CAPS_EMU_XML)
    xml = etree.tostring(doc)
    assert 'http://localhost:8094/wps' in xml
    xml = utils.replace_caps_url(xml, "https://localhost/ows/proxy/emu")
    assert 'http://localhost:8094/wps' not in xml
    assert 'https://localhost/ows/proxy/emu' in xml


def test_replace_caps_url_wms_111():
    doc = etree.parse(WMS_CAPS_NCWMS2_111_XML)
    xml = etree.tostring(doc)
    assert 'http://localhost:8080/ncWMS2/wms' in xml
    xml = utils.replace_caps_url(xml, "https://localhost/ows/proxy/wms")
    # assert 'http://localhost:8080/ncWMS2/wms' not in xml
    assert 'https://localhost/ows/proxy/wms' in xml


def test_replace_caps_url_wms_130():
    doc = etree.parse(WMS_CAPS_NCWMS2_130_XML)
    xml = etree.tostring(doc)
    assert 'http://localhost:8080/ncWMS2/wms' in xml
    xml = utils.replace_caps_url(xml, "https://localhost/ows/proxy/wms")
    # assert 'http://localhost:8080/ncWMS2/wms' not in xml
    assert 'https://localhost/ows/proxy/wms' in xml



"""
The owsproxy is based on `papyrus_ogcproxy <https://github.com/elemoine/papyrus_ogcproxy>`_

See also: https://github.com/nive/outpost/blob/master/outpost/proxy.py
"""

import urllib
import requests

from pyramid.response import Response
from pyramid.settings import asbool

from twitcher._compat import urlparse

from twitcher.owsexceptions import OWSAccessForbidden, OWSAccessFailed
from twitcher.utils import replace_caps_url
from twitcher.store import servicestore_factory


import logging
LOGGER = logging.getLogger(__name__)


allowed_content_types = (
    "application/xml",                       # XML
    "text/xml",
    "text/xml;charset=ISO-8859-1"
    "application/vnd.ogc.se_xml",            # OGC Service Exception
    "application/vnd.ogc.se+xml",            # OGC Service Exception
    # "application/vnd.ogc.success+xml",      # OGC Success (SLD Put)
    "application/vnd.ogc.wms_xml",           # WMS Capabilities
    # "application/vnd.ogc.gml",              # GML
    # "application/vnd.ogc.sld+xml",          # SLD
    "application/vnd.google-earth.kml+xml",  # KML
    "application/vnd.google-earth.kmz",
    "image/png",                             # PNG
    "image/png;mode=32bit",
    "image/gif",                             # GIF
    "image/jpeg",                            # JPEG
    "application/json",                      # JSON
    "application/json;charset=ISO-8859-1",
)

# TODO: configure allowed hosts
allowed_hosts = (
    # list allowed hosts here (no port limiting)
    # "localhost",
)


def _send_request(request, service, extra_path=None, request_params=None):

    # TODO: fix way to build url
    url = service['url']
    if extra_path:
        url += '/' + extra_path
    if service.get('c4i', False):
        if 'C4I-Access-Token' in request.headers:
            LOGGER.debug('using c4i token')
            url += '/' + request.headers['C4I-Access-Token']
    if request_params:
        url += '?' + request_params
    LOGGER.debug('url = %s', url)

    # forward request to target (without Host Header)
    h = dict(request.headers)
    h.pop("Host", h)
    try:
        resp = requests.request(method=request.method.upper(), url=url, data=request.body, headers=h)
    except Exception, e:
        return OWSAccessFailed("Request failed: {}".format(e.message))

    if resp.ok is False:
        if 'ExceptionReport' in resp.content:
            pass
        else:
            return OWSAccessFailed("Response is not ok: {}".format(resp.reason))

    # check for allowed content types
    ct = None
    # LOGGER.debug("headers=", resp.headers)
    if "Content-Type" in resp.headers:
        ct = resp.headers["Content-Type"]
        if not ct.split(";")[0] in allowed_content_types:
            msg = "Content type is not allowed: {}.".format(ct)
            LOGGER.error(msg)
            return OWSAccessForbidden(msg)
    else:
        # return OWSAccessFailed("Could not get content type from response.")
        LOGGER.warn("Could not get content type from response")

    try:
        if ct in ['text/xml', 'application/xml', 'text/xml;charset=ISO-8859-1']:
                # replace urls in xml content
                proxy_url = request.route_url('owsproxy', service_name=service['name'])
                # TODO: where do i need to replace urls?
                content = replace_caps_url(resp.content, proxy_url, service.get('url'))
        else:
            # raw content
            content = resp.content
    except:
        return OWSAccessFailed("Could not decode content.")

    headers = {}
    if ct:
        headers["Content-Type"] = ct
    return Response(content, status=resp.status_code, headers=headers)


def owsproxy_url(request):
    url = request.params.get("url")
    if url is None:
        return OWSAccessFailed("URL param is missing.")

    # check for full url
    parsed_url = urlparse(url)
    if not parsed_url.netloc or parsed_url.scheme not in ("http", "https"):
        return OWSAccessFailed("Not a valid URL.")
    return _send_request(request, service=dict(url=url, name='external'))


def owsproxy(request):
    """
    TODO: use ows exceptions
    """
    try:
        service_name = request.matchdict.get('service_name')
        extra_path = request.matchdict.get('extra_path')
        store = servicestore_factory(request.registry)
        service = store.fetch_by_name(service_name)
    except Exception as err:
        return OWSAccessFailed("Could not find service: {}.".format(err.message))
    else:
        return _send_request(request, service, extra_path, request_params=urllib.urlencode(request.params))


def owsproxy_delegate(request):
    """
    Delegates owsproxy request to external twitcher service.
    """
    twitcher_url = request.registry.settings.get('twitcher.url')
    url = twitcher_url + '/ows/proxy'
    if request.matchdict.get('service_name'):
        url += '/' + request.matchdict.get('service_name')
        if request.matchdict.get('access_token'):
            url += '/' + request.matchdict.get('service_name')
    url += '?' + urllib.urlencode(request.params)
    LOGGER.debug("delegate to owsproxy: %s", url)
    # forward request to target (without Host Header)
    # h = dict(request.headers)
    # h.pop("Host", h)
    resp = requests.request(method=request.method.upper(), url=url, data=request.body,
                            headers=request.headers, verify=False)
    return Response(resp.content, status=resp.status_code, headers=resp.headers)


def includeme(config):
    settings = config.registry.settings

    if asbool(settings.get('twitcher.ows_proxy', True)):
        LOGGER.debug('Twitcher /ows/proxy enabled.')

        config.add_route('owsproxy', '/ows/proxy/{service_name}')
        # TODO: maybe configure extra path
        # config.add_route('owsproxy_extra', '/ows/proxy/{service_name}/{extra_path:.*}')
        config.add_route('owsproxy_secured', '/ows/proxy/{service_name}/{access_token}')

        # use delegation mode?
        if asbool(settings.get('twitcher.ows_proxy_delegate', False)):
            LOGGER.debug('Twitcher /ows/proxy delegation mode enabled.')
            config.add_view(owsproxy_delegate, route_name='owsproxy')
            config.add_view(owsproxy_delegate, route_name='owsproxy_secured')
        else:
            # include twitcher config
            config.include('twitcher.config')
            # include mongodb
            config.include('twitcher.db')
            config.add_view(owsproxy, route_name='owsproxy')
            config.add_view(owsproxy, route_name='owsproxy_secured')
        # use /owsproxy?
        if asbool(settings.get('twitcher.ows_proxy_url', True)):
            LOGGER.debug('Twitcher /owsproxy enabled.')
            config.add_route('owsproxy_url', '/owsproxy')
            config.add_view(owsproxy_url, route_name='owsproxy_url')



from twitcher.namesgenerator import get_random_name
from twitcher.namesgenerator import get_sane_name


def test_get_random_name():
    name = get_random_name()
    assert len(name) > 3
    assert '_' in name


def test_get_random_name_retry():
    name = get_random_name(retry=True)
    assert len(name) > 3
    assert int(name[-1]) >= 0


def test_get_sane_name():
    assert get_sane_name("Hummingbird") == "hummingbird"
    assert get_sane_name("MapMint Demo Instance") == "mapmint_demo_instance"
    assert get_sane_name(None) is None
    assert get_sane_name("12") is None
    assert get_sane_name(" ab c ") == "ab_c"
    assert get_sane_name("a_much_to_long_name_for_this_test") == "a_much_to_long_name_for_t"



import os

from pywps import Process, LiteralInput, LiteralOutput

import logging
LOGGER = logging.getLogger("PYWPS")


class Hello(Process):
    def __init__(self):
        inputs = [
            LiteralInput('name', 'Your name', data_type='string')]
        outputs = [
            LiteralOutput('output', 'Output response',
                          data_type='string')]

        super(Hello, self).__init__(
            self._handler,
            identifier='hello',
            title='Say Hello',
            version='1.4',
            inputs=inputs,
            outputs=outputs,
            store_supported=True,
            status_supported=True
        )

    def _handler(self, request, response):
        response.update_status("saying hello ...", 0)
        LOGGER.debug("HOME=%s, Current Dir=%s", os.environ.get('HOME'), os.path.abspath(os.curdir))
        response.outputs['output'].data = 'Hello ' + request.inputs['name'][0].data
        return response



"""
Retrieve an ESGF certificate using a esgf access token and
prepare a `.dodsrc` file for OpenDAP_ access to the ESGF data archive.

This module uses code from esgf-slcs-client-example_ and esgf-pyclient_.


.. _esgf-slcs-client-example: https://github.com/cedadev/esgf-slcs-client-example
.. _esgf-pyclient: https://github.com/ESGF/esgf-pyclient
.. _OpenDAP: http://docs.opendap.org/index.php/Wiki_Testing/OPeNDAPUserGuideA

"""

import os
import shutil
from OpenSSL import crypto
import base64
import requests
from requests_oauthlib import OAuth2Session

from twitcher.utils import is_valid_url

import logging
logger = logging.getLogger(__name__)


ESGF_CERTS_DIR = 'certificates'
ESGF_CREDENTIALS = 'credentials.pem'
HTTP_RC = '.httprc'
DAP_CONFIG = '.dodsrc'
DAP_CONFIG_MARKER = '<<< Managed by twitcher >>>'

DAP_CONFIG_TEMPL = """\
# BEGIN {marker}
HTTP.VERBOSE={verbose}
HTTP.COOKIEJAR={base_dir}/.dods_cookies
HTTP.SSL.VALIDATE=0
HTTP.SSL.CERTIFICATE={esgf_credentials}
HTTP.SSL.KEY={esgf_credentials}
HTTP.SSL.CAPATH={esgf_certs_dir}
# END {marker}
"""


def fetch_certificate(workdir='.', data={}):
    try:
        url = data.get('esgf_slcs_service_url')
        access_token = data.get('esgf_access_token')
        test_credentials = data.get('esgf_credentials')
        mgr = ESGFAccessManager(url, base_dir=workdir)
        mgr.logon(access_token, test_credentials)
        logger.debug('Prepared twitcher workdir %s', workdir)
    except:
        logger.exception("Could not fetch certificate.")
        return False
    return True


class ESGFAccessManager(object):
    def __init__(self, slcs_service_url, base_dir=None):
        self.certificate_url = "{}/oauth/certificate/".format(slcs_service_url)
        self.base_dir = base_dir or '.'
        self.esgf_credentials = os.path.join(self.base_dir, ESGF_CREDENTIALS)
        self.esgf_certs_dir = os.path.join(self.base_dir, ESGF_CERTS_DIR)
        self.dap_config = os.path.join(self.base_dir, DAP_CONFIG)
        self.httprc = os.path.join(self.base_dir, HTTP_RC)

        os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'

    def logon(self, access_token, cert_url=None, timeout=1):
        if access_token:
            self._retrieve_certificate(access_token, timeout=timeout)
        elif cert_url:
            self._download_certificate(cert_url)
        else:
            raise Exception("Need either access_token or cert_url.")
        self._write_dap_config()
        # fix permission
        try:
            os.chmod(self.esgf_credentials, 0o400)
        except:
            logger.warn("Could not update permission of credentials.")
        return True

    def _download_certificate(self, url):
        if is_valid_url(url):
            logger.debug('Download cert from %s', url)
            response = requests.get(url, stream=True)
            with open(self.esgf_credentials, 'wb') as fd:
                for chunk in response.iter_content(chunk_size=128):
                    fd.write(chunk)
            return True
        else:
            return False

    def _retrieve_certificate(self, access_token, timeout=3):
        """
        Generates a new private key and certificate request, submits the request to be
        signed by the SLCS CA and returns the certificate.
        """
        logger.debug("Retrieve certificate with token.")

        # Generate a new key pair
        key_pair = crypto.PKey()
        key_pair.generate_key(crypto.TYPE_RSA, 2048)
        private_key = crypto.dump_privatekey(crypto.FILETYPE_PEM, key_pair).decode("utf-8")

        # Generate a certificate request using that key-pair
        cert_request = crypto.X509Req()

        # Create public key object
        cert_request.set_pubkey(key_pair)

        # Add the public key to the request
        cert_request.sign(key_pair, 'md5')
        der_cert_req = crypto.dump_certificate_request(crypto.FILETYPE_ASN1, cert_request)

        encoded_cert_req = base64.b64encode(der_cert_req)

        # Build the OAuth session object
        token = {'access_token': access_token, 'token_type': 'Bearer'}
        client = OAuth2Session(token=token)

        response = client.post(
            self.certificate_url,
            data={'certificate_request': encoded_cert_req},
            verify=False,
            timeout=timeout,
        )

        if response.ok:
            content = "{} {}".format(response.text, private_key)
            with open(self.esgf_credentials, 'w') as fh:
                fh.write(content)
            logger.debug('Fetched certificate successfully.')
        else:
            msg = "Could not get certificate: {} {}".format(response.status_code, response.reason)
            raise Exception(msg)
        return True

    def _write_dap_config(self, verbose=False, validate=False):
        content = DAP_CONFIG_TEMPL.format(
            verbose=1 if verbose else 0,
            validate=1 if validate else 0,
            base_dir=self.base_dir,
            esgf_certs_dir=self.esgf_certs_dir,
            esgf_credentials=self.esgf_credentials,
            marker=DAP_CONFIG_MARKER,
        )
        with open(self.dap_config, 'w') as fh:
            fh.write(content)
        shutil.copy2(self.dap_config, self.httprc)



"""
Definitions of types used by tokens.
"""

import time

from twitcher.utils import now_secs
from twitcher.exceptions import AccessTokenNotFound


import logging
logger = logging.getLogger(__name__)


class Service(dict):
    """
    Dictionary that contains OWS services. It always has ``'url'`` key.
    """
    def __init__(self, *args, **kwargs):
        super(Service, self).__init__(*args, **kwargs)
        if 'url' not in self:
            raise TypeError("'url' is required")

    @property
    def url(self):
        """Service URL."""
        return self['url']

    @property
    def name(self):
        """Service name."""
        return self.get('name', 'unknown')

    @property
    def type(self):
        """Service type."""
        return self.get('type', 'WPS')

    @property
    def public(self):
        """Flag if service has public access."""
        return self.get('public', False)

    @property
    def c4i(self):
        """Flag if service is by climate4impact."""
        return self.get('c4i', False)

    @property
    def params(self):
        return {'url': self.url, 'name': self.name, 'type': self.type, 'public': self.public, 'c4i': self.c4i}

    def __str__(self):
        return self.name

    def __repr__(self):
        cls = type(self)
        repr_ = dict.__repr__(self)
        return '{0}.{1}({2})'.format(cls.__module__, cls.__name__, repr_)


class AccessToken(dict):
    """
    Dictionary that contains access token. It always has ``'token'`` key.
    """

    def __init__(self, *args, **kwargs):
        super(AccessToken, self).__init__(*args, **kwargs)
        if 'token' not in self:
            raise TypeError("'token' is required")

    @property
    def token(self):
        """Access token string."""
        return self['token']

    @property
    def expires_at(self):
        return int(self.get("expires_at", 0))

    @property
    def expires_in(self):
        """
        Returns the time until the token expires.
        :return: The remaining time until expiration in seconds or 0 if the
                 token has expired.
        """
        time_left = self.expires_at - now_secs()

        if time_left > 0:
            return time_left
        return 0

    def is_expired(self):
        """
        Determines if the token has expired.
        :return: `True` if the token has expired. Otherwise `False`.
        """
        if self.expires_at is None:
            return True

        if self.expires_in > 0:
            return False

        return True

    @property
    def data(self):
        return self.get('data') or {}

    @property
    def params(self):
        return {'access_token': self.token, 'expires_at': self.expires_at}

    def __str__(self):
        return self.token

    def __repr__(self):
        cls = type(self)
        repr_ = dict.__repr__(self)
        return '{0}.{1}({2})'.format(cls.__module__, cls.__name__, repr_)



"""
Factories to create storage backends.
"""

# Interfaces
from twitcher.store.base import AccessTokenStore

# Factories
from twitcher.db import mongodb as _mongodb
from twitcher.store.mongodb import MongodbTokenStore
from twitcher.store.memory import MemoryTokenStore


def tokenstore_factory(registry, database=None):
    """
    Creates a token store with the interface of :class:`twitcher.store.AccessTokenStore`.
    By default the mongodb implementation will be used.

    :param database: A string with the store implementation name: "mongodb" or "memory".
    :return: An instance of :class:`twitcher.store.AccessTokenStore`.
    """
    database = database or 'mongodb'
    if database == 'mongodb':
        db = _mongodb(registry)
        store = MongodbTokenStore(db.tokens)
    else:
        store = MemoryTokenStore()
    return store


from twitcher.store.mongodb import MongodbServiceStore
from twitcher.store.memory import MemoryServiceStore


def servicestore_factory(registry, database=None):
    """
    Creates a service store with the interface of :class:`twitcher.store.ServiceStore`.
    By default the mongodb implementation will be used.

    :return: An instance of :class:`twitcher.store.ServiceStore`.
    """
    database = database or 'mongodb'
    if database == 'mongodb':
        db = _mongodb(registry)
        store = MongodbServiceStore(collection=db.services)
    else:
        store = MemoryServiceStore()
    return store



"""
Store adapters to persist and retrieve data during the twitcher process or
for later use. For example an access token storage and a service registry.

This module provides base classes that can be extended to implement your own
solution specific to your needs.

The implementation is based on `python-oauth2 <http://python-oauth2.readthedocs.io/en/latest/>`_.
"""


class AccessTokenStore(object):

    def save_token(self, access_token):
        """
        Stores an access token with additional data.
        """
        raise NotImplementedError

    def delete_token(self, token):
        """
        Deletes an access token from the store using its token string to identify it.
        This invalidates both the access token and the token.

        :param token: A string containing the token.
        :return: None.
        """
        raise NotImplementedError

    def fetch_by_token(self, token):
        """
        Fetches an access token from the store using its token string to
        identify it.

        :param token: A string containing the token.
        :return: An instance of :class:`twitcher.datatype.AccessToken`.
        """
        raise NotImplementedError

    def clear_tokens(self):
        """
        Removes all tokens from database.
        """
        raise NotImplementedError


class ServiceStore(object):
    """
    Storage for OWS services.
    """

    def save_service(self, service, overwrite=True):
        """
        Stores an OWS service with given name in storage.

        :param service: An instance of :class:`twitcher.datatype.Service`.
        """
        raise NotImplementedError

    def delete_service(self, name):
        """
        Removes service from database.
        """
        raise NotImplementedError

    def list_services(self):
        """
        Lists all services in database.
        """
        raise NotImplementedError

    def fetch_by_name(self, name):
        """
        Get service for given ``name`` from storage.

        :param token: A string containing the service name.
        :return: An instance of :class:`twitcher.datatype.Service`.
        """
        raise NotImplementedError

    def fetch_by_url(self, url):
        """
        Get service for given ``url`` from storage.

        :param token: A string containing the service url.
        :return: An instance of :class:`twitcher.datatype.Service`.
        """
        raise NotImplementedError

    def clear_services(self):
        """
        Removes all OWS services from storage.
        """
        raise NotImplementedError



"""
Based on tests from:

* https://github.com/geopython/pywps/tree/master/tests
* https://github.com/mmerickel/pyramid_services/tree/master/pyramid_services/tests
* http://webtest.pythonpaste.org/en/latest/
* http://docs.pylonsproject.org/projects/pyramid/en/latest/narr/testing.html
"""
import pytest
import unittest
import webtest
import pyramid.testing
from twitcher.tests.functional.common import setup_with_mongodb, setup_mongodb_tokenstore


class WpsAppTest(unittest.TestCase):

    def setUp(self):
        config = setup_with_mongodb()
        self.token = setup_mongodb_tokenstore(config)
        config.include('twitcher.wps')
        config.include('twitcher.tweens')
        self.app = webtest.TestApp(config.make_wsgi_app())

    def tearDown(self):
        pyramid.testing.tearDown()

    @pytest.mark.online
    def test_getcaps(self):
        resp = self.app.get('/ows/wps?service=wps&request=getcapabilities')
        assert resp.status_code == 200
        assert resp.content_type == 'text/xml'
        resp.mustcontain('</wps:Capabilities>')

    @pytest.mark.online
    def test_getcaps_with_invalid_token(self):
        resp = self.app.get('/ows/wps?service=wps&request=getcapabilities&access_token=invalid')
        assert resp.status_code == 200
        assert resp.content_type == 'text/xml'
        resp.mustcontain('</wps:Capabilities>')

    @pytest.mark.online
    def test_describeprocess(self):
        resp = self.app.get('/ows/wps?service=wps&request=describeprocess&version=1.0.0&identifier=hello')
        assert resp.status_code == 200
        assert resp.content_type == 'text/xml'
        resp.mustcontain('</wps:ProcessDescriptions>')

    @pytest.mark.online
    def test_describeprocess_with_invalid_token(self):
        resp = self.app.get(
            '/ows/wps?service=wps&request=describeprocess&version=1.0.0&identifier=hello&access_token=invalid')
        assert resp.status_code == 200
        assert resp.content_type == 'text/xml'
        resp.mustcontain('</wps:ProcessDescriptions>')

    @pytest.mark.online
    def test_execute_not_allowed(self):
        resp = self.app.get('/ows/wps?service=wps&request=execute&version=1.0.0&identifier=hello&datainputs=name=tux')
        assert resp.status_code == 200
        assert resp.content_type == 'text/xml'
        print resp.body
        resp.mustcontain('<Exception exceptionCode="NoApplicableCode" locator="AccessForbidden">')

    @pytest.mark.online
    def test_execute_allowed(self):
        url = "/ows/wps?service=wps&request=execute&version=1.0.0&identifier=hello&datainputs=name=tux"
        url += "&access_token={}".format(self.token)
        resp = self.app.get(url)
        assert resp.status_code == 200
        assert resp.content_type == 'text/xml'
        print resp.body
        resp.mustcontain(
            '<wps:ProcessSucceeded>PyWPS Process Say Hello finished</wps:ProcessSucceeded>')



from pyramid.view import view_config

import logging
logger = logging.getLogger(__name__)


@view_config(route_name='frontpage', renderer='json')
def frontpage(request):
    return {'message': 'hello'}


def includeme(config):
    config.add_route('frontpage', '/')



"""
Store adapters to read/write data to from/to mongodb using pymongo.
"""
import pymongo

from twitcher.store.base import AccessTokenStore
from twitcher.datatype import AccessToken
from twitcher.exceptions import AccessTokenNotFound

import logging
LOGGER = logging.getLogger(__name__)


class MongodbStore(object):
    """
    Base class extended by all concrete store adapters.
    """

    def __init__(self, collection):
        self.collection = collection


class MongodbTokenStore(AccessTokenStore, MongodbStore):
    def save_token(self, access_token):
        self.collection.insert_one(access_token)

    def delete_token(self, token):
        self.collection.delete_one({'token': token})

    def fetch_by_token(self, token):
        token = self.collection.find_one({'token': token})
        if not token:
            raise AccessTokenNotFound
        return AccessToken(token)

    def clear_tokens(self):
        self.collection.drop()


from twitcher.store.base import ServiceStore
from twitcher.datatype import Service
from twitcher.exceptions import ServiceRegistrationError
from twitcher.exceptions import ServiceNotFound
from twitcher import namesgenerator
from twitcher.utils import baseurl


class MongodbServiceStore(ServiceStore, MongodbStore):
    """
    Registry for OWS services. Uses mongodb to store service url and attributes.
    """

    def save_service(self, service, overwrite=True):
        """
        Stores an OWS service in mongodb.
        """

        service_url = baseurl(service.url)
        # check if service is already registered
        if self.collection.count({'url': service_url}) > 0:
            if overwrite:
                self.collection.delete_one({'url': service_url})
            else:
                raise ServiceRegistrationError("service url already registered.")

        name = namesgenerator.get_sane_name(service.name)
        if not name:
            name = namesgenerator.get_random_name()
            if self.collection.count({'name': name}) > 0:
                name = namesgenerator.get_random_name(retry=True)
        if self.collection.count({'name': name}) > 0:
            if overwrite:
                self.collection.delete_one({'name': name})
            else:
                raise Exception("service name already registered.")
        self.collection.insert_one(Service(
            url=service_url,
            name=name,
            type=service.type,
            public=service.public,
            c4i=service.c4i))
        return self.fetch_by_url(url=service_url)

    def delete_service(self, name):
        """
        Removes service from mongodb storage.
        """
        self.collection.delete_one({'name': name})
        return True

    def list_services(self):
        """
        Lists all services in mongodb storage.
        """
        my_services = []
        for service in self.collection.find().sort('name', pymongo.ASCENDING):
            my_services.append(Service(service))
        return my_services

    def fetch_by_name(self, name):
        """
        Gets service for given ``name`` from mongodb storage.
        """
        service = self.collection.find_one({'name': name})
        if not service:
            raise ServiceNotFound
        return Service(service)

    def fetch_by_url(self, url):
        """
        Gets service for given ``url`` from mongodb storage.
        """
        service = self.collection.find_one({'url': baseurl(url)})
        if not service:
            raise ServiceNotFound
        return Service(service)

    def clear_services(self):
        """
        Removes all OWS services from mongodb storage.
        """
        self.collection.drop()
        return True



"""
Based on unitests in https://github.com/wndhydrnt/python-oauth2/tree/master/oauth2/test
"""

import pytest
import unittest
import mock

from twitcher.datatype import AccessToken
from twitcher.datatype import Service
from twitcher.utils import expires_at


class AccessTokenTestCase(unittest.TestCase):

    def test_access_token(self):
        access_token = AccessToken(token='abcdef', expires_at=expires_at(hours=1))
        assert access_token.token == 'abcdef'
        assert access_token.expires_in > 0
        assert access_token.expires_in <= 3600
        assert access_token.is_expired() is False
        assert access_token.params['access_token'] == 'abcdef'
        assert 'expires_at' in access_token.params

    def test_missing_token(self):
        with pytest.raises(TypeError) as e_info:
            AccessToken()

    def test_invalid_access_token(self):
        access_token = AccessToken(token='abcdef', expires_at=expires_at(hours=-1))
        assert access_token.expires_in == 0
        assert access_token.is_expired() is True

    def test_access_token_with_data(self):
        access_token = AccessToken(token='12345', expires_at=expires_at(hours=1),
                                   data={'esgf_token': 'bfghk'})
        assert access_token.data == {'esgf_token': 'bfghk'}


class ServiceTestCase(unittest.TestCase):
    def test_service_with_url_only(self):
        service = Service(url='http://nowhere/wps')
        assert service.url == 'http://nowhere/wps'
        assert service.name == 'unknown'

    def test_missing_url(self):
        with pytest.raises(TypeError) as e_info:
            Service()

    def test_service_with_name(self):
        service = Service(url='http://nowhere/wps', name="test_wps")
        assert service.url == 'http://nowhere/wps'
        assert service.name == 'test_wps'

    def test_service_params(self):
        service = Service(url='http://nowhere/wps', name="test_wps")
        assert service.params == {'c4i': False,
                                  'name': 'test_wps',
                                  'public': False,
                                  'type': 'WPS',
                                  'url': 'http://nowhere/wps'}



from pyramid.view import view_defaults
from pyramid_rpc.xmlrpc import xmlrpc_method
from pyramid.settings import asbool

from twitcher.api import ITokenManager, TokenManager
from twitcher.api import IRegistry, Registry
from twitcher.tokengenerator import tokengenerator_factory
from twitcher.store import tokenstore_factory
from twitcher.store import servicestore_factory

import logging
LOGGER = logging.getLogger("TWITCHER")


@view_defaults(permission='view')
class RPCInterface(ITokenManager, IRegistry):
    def __init__(self, request):
        self.request = request
        self.tokenmgr = TokenManager(
            tokengenerator_factory(request.registry),
            tokenstore_factory(request.registry))
        self.srvreg = Registry(servicestore_factory(request.registry))

    def generate_token(self, valid_in_hours=1, environ=None):
        """
        Implementation of :meth:`twitcher.api.ITokenManager.generate_token`.
        """
        return self.tokenmgr.generate_token(valid_in_hours, environ)

    def revoke_token(self, token):
        """
        Implementation of :meth:`twitcher.api.ITokenManager.revoke_token`.
        """
        return self.tokenmgr.revoke_token(token)

    def revoke_all_tokens(self):
        """
        Implementation of :meth:`twitcher.api.ITokenManager.revoke_all_tokens`.
        """
        return self.tokenmgr.revoke_all_tokens()

    def register_service(self, url, data=None, overwrite=True):
        """
        Implementation of :meth:`twitcher.api.IRegistry.register_service`.
        """
        return self.srvreg.register_service(url, data, overwrite)

    def unregister_service(self, name):
        """
        Implementation of :meth:`twitcher.api.IRegistry.unregister_service`.
        """
        return self.srvreg.unregister_service(name)

    def get_service_by_name(self, name):
        """
        Implementation of :meth:`twitcher.api.IRegistry.get_service_by_name`.
        """
        return self.srvreg.get_service_by_name(name)

    def get_service_by_url(self, url):
        """
        Implementation of :meth:`twitcher.api.IRegistry.get_service_by_url`.
        """
        return self.srvreg.get_service_by_url(url)

    def list_services(self):
        """
        Implementation of :meth:`twitcher.api.IRegistry.list_services`.
        """
        return self.srvreg.list_services()

    def clear_services(self):
        """
        Implementation of :meth:`twitcher.api.IRegistry.clear_services`.
        """
        return self.srvreg.clear_services()


def includeme(config):
    """ The callable makes it possible to include rpcinterface
    in a Pyramid application.

    Calling ``config.include(twitcher.rpcinterface)`` will result in this
    callable being called.

    Arguments:

    * ``config``: the ``pyramid.config.Configurator`` object.
    """
    settings = config.registry.settings

    if asbool(settings.get('twitcher.rpcinterface', True)):
        LOGGER.debug('Twitcher XML-RPC Interface enabled.')

        # include twitcher config
        config.include('twitcher.config')

        # using basic auth
        config.include('twitcher.basicauth')

        # pyramid xml-rpc
        # http://docs.pylonsproject.org/projects/pyramid-rpc/en/latest/xmlrpc.html
        config.include('pyramid_rpc.xmlrpc')
        config.include('twitcher.db')
        config.add_xmlrpc_endpoint('api', '/RPC2')

        # register xmlrpc methods
        config.add_xmlrpc_method(RPCInterface, attr='generate_token', endpoint='api', method='generate_token')
        config.add_xmlrpc_method(RPCInterface, attr='revoke_token', endpoint='api', method='revoke_token')
        config.add_xmlrpc_method(RPCInterface, attr='revoke_all_tokens', endpoint='api', method='revoke_all_tokens')
        config.add_xmlrpc_method(RPCInterface, attr='register_service', endpoint='api', method='register_service')
        config.add_xmlrpc_method(RPCInterface, attr='unregister_service', endpoint='api', method='unregister_service')
        config.add_xmlrpc_method(RPCInterface, attr='get_service_by_name', endpoint='api', method='get_service_by_name')
        config.add_xmlrpc_method(RPCInterface, attr='get_service_by_url', endpoint='api', method='get_service_by_url')
        config.add_xmlrpc_method(RPCInterface, attr='clear_services', endpoint='api', method='clear_services')
        config.add_xmlrpc_method(RPCInterface, attr='list_services', endpoint='api', method='list_services')



import tempfile

from twitcher.exceptions import AccessTokenNotFound
from twitcher.exceptions import ServiceNotFound
from twitcher.owsexceptions import OWSAccessForbidden, OWSInvalidParameterValue
from twitcher.utils import path_elements
from twitcher.store import tokenstore_factory
from twitcher.store import servicestore_factory
from twitcher.utils import parse_service_name
from twitcher.owsrequest import OWSRequest
from twitcher.esgf import fetch_certificate, ESGF_CREDENTIALS

import logging
LOGGER = logging.getLogger("TWITCHER")

protected_path = '/ows/'


def owssecurity_factory(registry):
    return OWSSecurity(tokenstore_factory(registry), servicestore_factory(registry))


class OWSSecurity(object):

    def __init__(self, tokenstore, servicestore):
        self.tokenstore = tokenstore
        self.servicestore = servicestore

    def get_token_param(self, request):
        token = None
        if 'token' in request.params:
            token = request.params['token']   # in params
        elif 'access_token' in request.params:
            token = request.params['access_token']   # in params
        elif 'Access-Token' in request.headers:
            token = request.headers['Access-Token']  # in header
        else:  # in path
            elements = path_elements(request.path)
            if len(elements) > 1:  # there is always /ows/
                token = elements[-1]   # last path element
        return token

    def prepare_headers(self, request, access_token):
        if "esgf_access_token" in access_token.data or "esgf_credentials" in access_token.data:
            workdir = tempfile.mkdtemp(prefix=request.prefix, dir=request.workdir)
            if fetch_certificate(workdir=workdir, data=access_token.data):
                request.headers['X-Requested-Workdir'] = workdir
                request.headers['X-X509-User-Proxy'] = workdir + '/' + ESGF_CREDENTIALS
                LOGGER.debug("Prepared request headers.")
        return request

    def check_request(self, request):
        if request.path.startswith(protected_path):
            # TODO: fix this code
            try:
                service_name = parse_service_name(request.path)
                service = self.servicestore.fetch_by_name(service_name)
                is_public = service.public
                if service.public is True:
                    LOGGER.warn('public access for service %s', service_name)
            except ServiceNotFound:
                is_public = False
                LOGGER.warn("Service not registered.")
            ows_request = OWSRequest(request)
            if not ows_request.service_allowed():
                raise OWSInvalidParameterValue(
                    "service %s not supported" % ows_request.service, value="service")
            if not ows_request.public_access():
                try:
                    # try to get access_token ... if no access restrictions then don't complain.
                    token = self.get_token_param(request)
                    access_token = self.tokenstore.fetch_by_token(token)
                    if access_token.is_expired() and not is_public:
                        raise OWSAccessForbidden("Access token is expired.")
                    # update request with data from access token
                    # request.environ.update(access_token.data)
                    request = self.prepare_headers(request, access_token)
                except AccessTokenNotFound:
                    if not is_public:
                        raise OWSAccessForbidden("Access token is required to access this service.")



import pytest
import unittest
import mock

from pyramid.testing import DummyRequest

from twitcher.datatype import AccessToken
from twitcher.datatype import Service
from twitcher.utils import expires_at
from twitcher.owssecurity import OWSSecurity
from twitcher.owsexceptions import OWSAccessForbidden
from twitcher.store.memory import MemoryTokenStore
from twitcher.store.memory import MemoryServiceStore


class OWSSecurityTestCase(unittest.TestCase):
    def setUp(self):
        self.access_token = AccessToken(token="cdefg", expires_at=expires_at(hours=1))

        self.tokenstore = MemoryTokenStore()
        self.tokenstore.save_token(self.access_token)
        self.empty_tokenstore = MemoryTokenStore()

        self.service = Service(url='http://nowhere/wps', name='test_wps', public=False)
        self.servicestore = MemoryServiceStore()
        self.servicestore.save_service(self.service)

        self.security = OWSSecurity(tokenstore=self.tokenstore, servicestore=self.servicestore)

    def test_get_token_by_param(self):
        params = dict(request="Execute", service="WPS", access_token="abcdef")
        request = DummyRequest(params=params)
        token = self.security.get_token_param(request)
        assert token == "abcdef"

    def test_get_token_by_path(self):
        params = dict(request="Execute", service="WPS")
        request = DummyRequest(params=params, path="/ows/proxy/emu/12345")
        token = self.security.get_token_param(request)
        assert token == "12345"

    def test_get_token_by_header(self):
        params = dict(request="Execute", service="WPS")
        headers = {'Access-Token': '54321'}
        request = DummyRequest(params=params, headers=headers)
        token = self.security.get_token_param(request)
        assert token == "54321"

    def test_check_request(self):
        params = dict(request="Execute", service="WPS", version="1.0.0", token="cdefg")
        request = DummyRequest(params=params, path='/ows/proxy/emu')
        self.security.check_request(request)

    def test_check_request_invalid(self):
        security = OWSSecurity(tokenstore=self.empty_tokenstore, servicestore=self.servicestore)

        params = dict(request="Execute", service="WPS", version="1.0.0", token="xyz")
        request = DummyRequest(params=params, path='/ows/proxy/emu')
        with pytest.raises(OWSAccessForbidden) as e_info:
            security.check_request(request)

    def test_check_request_allowed_caps(self):
        security = OWSSecurity(tokenstore=self.empty_tokenstore, servicestore=self.servicestore)

        params = dict(request="GetCapabilities", service="WPS", version="1.0.0")
        request = DummyRequest(params=params, path='/ows/proxy/emu')
        security.check_request(request)

    def test_check_request_allowed_describeprocess(self):
        security = OWSSecurity(tokenstore=self.empty_tokenstore, servicestore=self.servicestore)

        params = dict(request="DescribeProcess", service="WPS", version="1.0.0")
        request = DummyRequest(params=params, path='/ows/proxy/emu')
        security.check_request(request)

    def test_check_request_public_access(self):
        servicestore = MemoryServiceStore()
        servicestore.save_service(Service(
            url='http://nowhere/wps', name='test_wps', public=True))
        security = OWSSecurity(tokenstore=self.tokenstore, servicestore=servicestore)

        params = dict(request="Execute", service="WPS", version="1.0.0", token="cdefg")
        request = DummyRequest(params=params, path='/ows/proxy/emu')
        security.check_request(request)



"""
Read or write data from or to local memory.

Though not very valuable in a production setup, these store adapters are great
for testing purposes.
"""

from twitcher.store.base import AccessTokenStore
from twitcher.exceptions import AccessTokenNotFound


class MemoryTokenStore(AccessTokenStore):
    """
    Stores tokens in memory.
    Useful for testing purposes or APIs with a very limited set of clients.

    Use mongodb as storage to be able to scale.
    """
    def __init__(self):
        self.access_tokens = {}

    def save_token(self, access_token):
        self.access_tokens[access_token.token] = access_token
        return True

    def delete_token(self, token):
        if token in self.access_tokens:
            del self.access_tokens[token]

    def fetch_by_token(self, token):
        if token not in self.access_tokens:
            raise AccessTokenNotFound

        return self.access_tokens[token]

    def clear_tokens(self):
        self.access_tokens = {}


from twitcher.store.base import ServiceStore
from twitcher.datatype import Service
from twitcher.exceptions import ServiceRegistrationError
from twitcher.exceptions import ServiceNotFound
from twitcher import namesgenerator
from twitcher.utils import baseurl


class MemoryServiceStore(ServiceStore):
    """
    Stores OWS services in memory. Useful for testing purposes.
    """
    def __init__(self):
        self.url_index = {}
        self.name_index = {}

    def _delete(self, url=None, name=None):
        if url:
            service = self.url_index[url]
            del self.name_index[service['name']]
            del self.url_index[url]
        elif name:
            service = self.name_index[name]
            del self.url_index[service['url']]
            del self.name_index[name]

    def _insert(self, service):
        self.name_index[service['name']] = service
        self.url_index[service['url']] = service

    def save_service(self, service, overwrite=True):
        """
        Store an OWS service in database.
        """

        service_url = baseurl(service.url)
        # check if service is already registered
        if service_url in self.url_index:
            if overwrite:
                self._delete(url=service_url)
            else:
                raise ServiceRegistrationError("service url already registered.")

        name = namesgenerator.get_sane_name(service.name)
        if not name:
            name = namesgenerator.get_random_name()
            if name in self.name_index:
                name = namesgenerator.get_random_name(retry=True)
        if name in self.name_index:
            if overwrite:
                self._delete(name=name)
            else:
                raise Exception("service name already registered.")
        self._insert(Service(
            url=service_url,
            name=name,
            type=service.type,
            public=service.public,
            c4i=service.c4i))
        return self.fetch_by_url(url=service_url)

    def delete_service(self, name):
        """
        Removes service from registry database.
        """
        self._delete(name=name)
        return True

    def list_services(self):
        """
        Lists all services in memory storage.
        """
        my_services = []
        for service in self.url_index.itervalues():
            my_services.append(Service(service))
        return my_services

    def fetch_by_name(self, name):
        """
        Get service for given ``name`` from memory storage.
        """
        service = self.name_index.get(name)
        if not service:
            raise ServiceNotFound
        return Service(service)

    def fetch_by_url(self, url):
        """
        Get service for given ``url`` from memory storage.
        """
        service = self.url_index.get(baseurl(url))
        if not service:
            raise ServiceNotFound
        return Service(service)

    def clear_services(self):
        """
        Removes all OWS services from memory storage.
        """
        self.url_index = {}
        self.name_index = {}
        return True



import time
from datetime import datetime
import pytz
from lxml import etree

from twitcher.exceptions import ServiceNotFound

from twitcher._compat import urlparse

import logging
logger = logging.getLogger(__name__)


def is_valid_url(url):
    try:
        parsed_url = urlparse(url)
        return True if all([parsed_url.scheme, ]) else False
    except:
        return False


def parse_service_name(url):
    parsed_url = urlparse(url)
    service_name = None
    if parsed_url.path.startswith("/ows/proxy"):
        parts = parsed_url.path.strip('/').split('/')
        if len(parts) > 2:
            service_name = parts[2]
    if not service_name:
        raise ServiceNotFound
    return service_name


def now():
    return localize_datetime(datetime.utcnow())


def now_secs():
    """
    Return the current time in seconds since the Epoch.
    """
    return int(time.time())


def expires_at(hours=1):
    return now_secs() + hours * 3600


def localize_datetime(dt, tz_name='UTC'):
    """Provide a timzeone-aware object for a given datetime and timezone name
    """
    tz_aware_dt = dt
    if dt.tzinfo is None:
        utc = pytz.timezone('UTC')
        aware = utc.localize(dt)
        timezone = pytz.timezone(tz_name)
        tz_aware_dt = aware.astimezone(timezone)
    else:
        logger.warn('tzinfo already set')
    return tz_aware_dt


def baseurl(url):
    """
    return baseurl of given url
    """
    parsed_url = urlparse(url)
    if not parsed_url.netloc or parsed_url.scheme not in ("http", "https"):
        raise ValueError('bad url')
    service_url = "%s://%s%s" % (parsed_url.scheme, parsed_url.netloc, parsed_url.path.strip())
    return service_url


def path_elements(path):
    elements = [el.strip() for el in path.split('/')]
    elements = [el for el in elements if len(el) > 0]
    return elements


def lxml_strip_ns(tree):
    for node in tree.iter():
        try:
            has_namespace = node.tag.startswith('{')
        except AttributeError:
            continue  # node.tag is not a string (node is a comment or similar)
        if has_namespace:
            node.tag = node.tag.split('}', 1)[1]


def replace_caps_url(xml, url, prev_url=None):
    ns = {
        'ows': 'http://www.opengis.net/ows/1.1',
        'xlink': 'http://www.w3.org/1999/xlink'}
    doc = etree.fromstring(xml)
    # wms 1.1.1 onlineResource
    if 'WMT_MS_Capabilities' in doc.tag:
        logger.debug("replace proxy urls in wms 1.1.1")
        for element in doc.findall('.//OnlineResource[@xlink:href]', namespaces=ns):
            parsed_url = urlparse(element.get('{http://www.w3.org/1999/xlink}href'))
            new_url = url
            if parsed_url.query:
                new_url += '?' + parsed_url.query
            element.set('{http://www.w3.org/1999/xlink}href', new_url)
        xml = etree.tostring(doc)
    # wms 1.3.0 onlineResource
    elif 'WMS_Capabilities' in doc.tag:
        logger.debug("replace proxy urls in wms 1.3.0")
        for element in doc.findall('.//{http://www.opengis.net/wms}OnlineResource[@xlink:href]', namespaces=ns):
            parsed_url = urlparse(element.get('{http://www.w3.org/1999/xlink}href'))
            new_url = url
            if parsed_url.query:
                new_url += '?' + parsed_url.query
            element.set('{http://www.w3.org/1999/xlink}href', new_url)
        xml = etree.tostring(doc)
    # wps operations
    elif 'Capabilities' in doc.tag:
        for element in doc.findall('ows:OperationsMetadata//*[@xlink:href]', namespaces=ns):
            element.set('{http://www.w3.org/1999/xlink}href', url)
        xml = etree.tostring(doc)
    elif prev_url:
        xml = xml.decode('utf-8', 'ignore')
        xml = xml.replace(prev_url, url)
    return xml



"""
based on `papyrus_ogcproxy <https://github.com/elemoine/papyrus_ogcproxy>`_

pyramid testing:

* http://docs.pylonsproject.org/projects/pyramid/en/latest/quick_tutorial/routing.html
"""

import unittest

from pyramid import testing
from pyramid.testing import DummyRequest

from twitcher.owsexceptions import OWSAccessFailed
from twitcher import owsproxy
from twitcher.owsproxy import owsproxy as owsproxy_view


class OWSProxyTests(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp()

    def tearDown(self):
        testing.tearDown()
        owsproxy.allowed_hosts = ()

    def test_badrequest_url(self):
        request = DummyRequest(scheme='http')
        response = owsproxy_view(request)
        assert isinstance(response, OWSAccessFailed) is True

    def test_badrequest_netloc(self):
        request = DummyRequest(scheme='http',
                               params={'url': 'http://'})
        response = owsproxy_view(request)
        assert isinstance(response, OWSAccessFailed) is True



import sys
import getpass
import argcomplete
import argparse

from twitcher.client import TwitcherService

import logging
logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.WARN)
LOGGER = logging.getLogger(__name__)


class TwitcherCtl(object):
    """
    Command line to interact with the xmlrpc interface of the ``twitcher`` service.
    """

    def create_parser(self):
        parser = argparse.ArgumentParser(
            prog="twitcherctl",
            description='twitcherctl -- control twitcher service from the cmd line.',
        )
        parser.add_argument("--debug",
                            help="Enable debug mode.",
                            action="store_true")
        parser.add_argument('-s', '--serverurl',
                            metavar='URL',
                            default='https://localhost:5000',
                            help='URL on which twitcher server is listening (default "https://localhost:5000").')
        parser.add_argument("-u", "--username",
                            help="Username to use for authentication with server.")
        parser.add_argument("-p", "--password",
                            help="Password to use for authentication with server")
        parser.add_argument("-k", "--insecure",  # like curl
                            help="Don't validate the server's certificate.",
                            action="store_true")

        # commands
        subparsers = parser.add_subparsers(
            dest='cmd',
            title='command',
            description='List of available commands',
        )

        # token managment
        # ---------------

        # gentoken
        subparser = subparsers.add_parser('gentoken', help="Generates an access token.")
        subparser.add_argument('-H', '--valid-in-hours', type=int, default=1,
                               help="Set how long the token is valid in hours (default: 1 hour).")
        subparser.add_argument('-S', '--esgf-slcs-service-url', default="https://172.28.128.3",
                               help="URL of ESGF SLCS service (default: https://172.28.128.3).")
        subparser.add_argument('-T', '--esgf-access-token',
                               help="ESGF access token to retrieve a certificate from ESGF SLCS service.")
        subparser.add_argument('-C', '--esgf-credentials',
                               help="URL pointing to ESGF credentials.")
        subparser.add_argument('-e', '--env', nargs='*', default=[],
                               help="Set environment variable (key=value).")

        # revoke
        subparser = subparsers.add_parser('revoke', help="Remove given access token.")
        subparser.add_argument('token', nargs="?", help="access token")
        subparser.add_argument('-A', '--all', action="store_true",
                               help="Remove all access tokens.")

        # service registry
        # ----------------

        # list
        subparser = subparsers.add_parser('list', help="Lists all registered OWS services used by OWS proxy.")

        # clear
        subparser = subparsers.add_parser('clear', help="Removes all OWS services from the registry.")

        # register
        subparser = subparsers.add_parser('register',
                                          help="Adds OWS service to the registry to be used by the OWS proxy.")
        subparser.add_argument('url', help="Service url.")
        subparser.add_argument('--name', help="Service name. If not set then a name will be generated.")
        subparser.add_argument('--type', default='wps',
                               help="Service type (wps, wms). Default: wps.")
        subparser.add_argument('--public', action='store_true',
                               help="If set then service has no access restrictions.")

        # unregister
        subparser = subparsers.add_parser('unregister', help="Removes OWS service from the registry.")
        subparser.add_argument('name', help="Service name.")

        return parser

    def run(self, args):
        if args.debug:
            LOGGER.setLevel(logging.DEBUG)

        if args.insecure:
            LOGGER.warn('disabled certificate verification!')

        password = args.password
        if args.username:
            # username = raw_input('Username:')
            if not password:
                password = getpass.getpass(prompt='Password:')

        verify_ssl = args.insecure is False
        service = TwitcherService(url=args.serverurl,
                                  username=args.username,
                                  password=password,
                                  verify=verify_ssl)
        result = None
        try:
            if args.cmd == 'list':
                result = service.list_services()
            elif args.cmd == 'register':
                result = service.register_service(
                    url=args.url,
                    data={'name': args.name, 'type': args.type, 'public': args.public})
            elif args.cmd == 'unregister':
                result = service.unregister_service(name=args.name)
            elif args.cmd == 'clear':
                result = service.clear_services()
            elif args.cmd == 'gentoken':
                data = {k: v for k, v in (x.split('=') for x in args.env)}
                if args.esgf_access_token:
                    data['esgf_access_token'] = args.esgf_access_token
                    data['esgf_slcs_service_url'] = args.esgf_slcs_service_url
                if args.esgf_credentials:
                    data['esgf_credentials'] = args.esgf_credentials
                result = service.generate_token(valid_in_hours=args.valid_in_hours, data=data)
            elif args.cmd == 'revoke':
                if args.all is True:
                    result = service.revoke_all_tokens()
                else:
                    result = service.revoke_token(token=args.token)
        except Exception as e:
            LOGGER.error("Error: %s", e.message)
        else:
            LOGGER.info("Result: %s", result)
            return result


def main():
    LOGGER.setLevel(logging.INFO)

    ctl = TwitcherCtl()
    parser = ctl.create_parser()
    argcomplete.autocomplete(parser)

    args = parser.parse_args()
    return ctl.run(args)


if __name__ == '__main__':
    sys.exit(main())



"""
Provides various implementations of algorithms to generate an Access Token.
"""

import uuid

from twitcher.datatype import AccessToken
from twitcher.utils import expires_at

import logging
logger = logging.getLogger(__name__)


def tokengenerator_factory(registry):
    return UuidTokenGenerator()


class TokenGenerator(object):
    """
    Base class for every token generator.
    """
    def create_access_token(self, valid_in_hours=1, data=None):
        """
        Creates an access token.

        TODO: check valid in hours
        TODO: maybe specify how often a token can be used
        """
        data = data or {}
        token = AccessToken(
            token=self.generate(),
            expires_at=expires_at(hours=valid_in_hours),
            data=data)
        return token

    def generate(self):
        raise NotImplementedError


class UuidTokenGenerator(TokenGenerator):
    """
    Generate a token using uuid4.
    """
    def generate(self):
        """
        :return: A new token
        """
        return uuid.uuid4().get_hex()



from twitcher.datatype import Service

import logging
LOGGER = logging.getLogger("TWITCHER")


class ITokenManager(object):
    def generate_token(self, valid_in_hours=1, data=None):
        """
        Generates an access token which is valid for ``valid_in_hours``.

        Arguments:

        * :param valid_in_hours: an int with number of hours the token is valid.
        * :param data: a dict with extra data used with this token.

        Possible keys: ``esgf_access_token``, ``esgf_slcs_service_url`` or ``esgf_credentials``.
        """
        raise NotImplementedError

    def revoke_token(self, token):
        """
        Remove token from tokenstore.
        """
        raise NotImplementedError

    def revoke_all_tokens(self):
        """
        Removes all tokens from tokenstore.
        """
        raise NotImplementedError


class IRegistry(object):
    def register_service(self, url, data, overwrite):
        """
        Adds an OWS service with the given ``url`` to the service store.

        :param data: a dict with additional information like ``name``.
        """
        raise NotImplementedError

    def unregister_service(self, name):
        """
        Removes OWS service with the given ``name`` from the service store.
        """
        raise NotImplementedError

    def get_service_by_name(self, name):
        """
        Gets service with given ``name`` from service store.
        """
        raise NotImplementedError

    def get_service_by_url(self, url):
        """
        Gets service with given ``url`` from service store.
        """
        raise NotImplementedError

    def list_services(self):
        """
        Lists all registred OWS services.
        """
        raise NotImplementedError

    def clear_services(self):
        """
        Removes all services from the service store.
        """
        raise NotImplementedError


class TokenManager(ITokenManager):
    """
    Implementation of :class:`twitcher.api.ITokenManager`.
    """

    def __init__(self, tokengenerator, tokenstore):
        self.tokengenerator = tokengenerator
        self.store = tokenstore

    def generate_token(self, valid_in_hours=1, data=None):
        """
        Implementation of :meth:`twitcher.api.ITokenManager.generate_token`.
        """
        data = data or {}
        access_token = self.tokengenerator.create_access_token(
            valid_in_hours=valid_in_hours,
            data=data,
        )
        self.store.save_token(access_token)
        return access_token.params

    def revoke_token(self, token):
        """
        Implementation of :meth:`twitcher.api.ITokenManager.revoke_token`.
        """
        try:
            self.store.delete_token(token)
        except:
            LOGGER.exception('Failed to remove token.')
            return False
        else:
            return True

    def revoke_all_tokens(self):
        """
        Implementation of :meth:`twitcher.api.ITokenManager.revoke_all_tokens`.
        """
        try:
            self.store.clear_tokens()
        except:
            LOGGER.exception('Failed to remove tokens.')
            return False
        else:
            return True


class Registry(IRegistry):
    """
    Implementation of :class:`twitcher.api.IRegistry`.
    """
    def __init__(self, servicestore):
        self.store = servicestore

    def register_service(self, url, data=None, overwrite=True):
        """
        Implementation of :meth:`twitcher.api.IRegistry.register_service`.
        """
        data = data or {}

        args = dict(data)
        args['url'] = url
        service = Service(**args)
        service = self.store.save_service(service, overwrite=overwrite)
        return service.params

    def unregister_service(self, name):
        """
        Implementation of :meth:`twitcher.api.IRegistry.unregister_service`.
        """
        try:
            self.store.delete_service(name=name)
        except:
            LOGGER.exception('unregister failed')
            return False
        else:
            return True

    def get_service_by_name(self, name):
        """
        Implementation of :meth:`twitcher.api.IRegistry.get_service_by_name`.
        """
        try:
            service = self.store.fetch_by_name(name=name)
        except:
            LOGGER.error('Could not get service with name %s', name)
            return {}
        else:
            return service.params

    def get_service_by_url(self, url):
        """
        Implementation of :meth:`twitcher.api.IRegistry.get_service_by_url`.
        """
        try:
            service = self.store.fetch_by_url(url=url)
        except:
            LOGGER.error('Could not get service with url %s', url)
            return {}
        else:
            return service.params

    def list_services(self):
        """
        Implementation of :meth:`twitcher.api.IRegistry.list_services`.
        """
        try:
            services = [service.params for service in self.store.list_services()]
        except:
            LOGGER.error('List services failed.')
            return []
        else:
            return services

    def clear_services(self):
        """
        Implementation of :meth:`twitcher.api.IRegistry.clear_services`.
        """
        try:
            self.store.clear_services()
        except:
            LOGGER.error('Clear services failed.')
            return False
        else:
            return True



from pyramid.httpexceptions import HTTPUnauthorized
from pyramid.security import forget
from pyramid.view import forbidden_view_config
from pyramid.authentication import BasicAuthAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.security import (
    Allow,
    Everyone,
    ALL_PERMISSIONS)

import logging
logger = logging.getLogger(__name__)


Admin = 'group:admin'


def groupfinder(username, password, request):
    logger.error("u=%s, p=%s", request.username, request.password)
    if request.username and request.password:
        if username == request.username and password == request.password:
            return [Admin]
        else:
            return []
    else:
        # Warning: no access restrictions!
        return [Admin]


class RootFactory(object):
    __acl__ = [
        (Allow, Everyone, 'view'),
        (Allow, Admin, ALL_PERMISSIONS)
    ]

    def __init__(self, request):
        pass


def root_factory(request):
    return RootFactory(request)


@forbidden_view_config()
def basic_challenge(request):
    response = HTTPUnauthorized()
    response.headers.update(forget(request))
    return response


def _get_username(request):
    settings = request.registry.settings
    if 'twitcher.username' in settings:
        username = settings['twitcher.username']
        if username:
            username = username.strip()
            if len(username) > 2:
                return username
    return None


def _get_password(request):
    settings = request.registry.settings
    if 'twitcher.password' in settings:
        password = settings['twitcher.password']
        if password:
            password = password.strip()
            if len(password) > 2:
                return password
    return None


def auth_activated(registry):
    settings = registry.settings
    username = settings.get('twitcher.username')
    if username:
        if len(username.strip()) > 2:
            return True
    return False


def includeme(config):
    if auth_activated(config.registry):
        logger.debug("basic authentication is activated.")
        # Security policies for basic auth
        authn_policy = BasicAuthAuthenticationPolicy(check=groupfinder, realm="Birdhouse")
        authz_policy = ACLAuthorizationPolicy()
        config.set_authentication_policy(authn_policy)
        config.set_authorization_policy(authz_policy)
        config.set_root_factory(root_factory)
        config.add_request_method(_get_username, 'username', reify=True)
        config.add_request_method(_get_password, 'password', reify=True)
